#=============================================================================#
# vtksuite                                                                    #
#                                                                             #
#                                                                             #
#                                                                             #
# Authors: Jacopo Canton, Nicolo' Fabbiane                                    #
# Contacts: jcanton(at)mech.kth.se, nicolo(at)mech.kth.se                     #
# Last edit: 2016-01-28                                                       #
#=============================================================================#
import numpy as np
from tvtk.api import tvtk


#==============================================================================
def exa2vtk(field, downsample):
	"""A function for converting exadata to vtk data

	Parameters
	----------
	field : str
		a dataset in nekdata format
	downsample : bool
		flag T/F
	"""
	#
	if (downsample):
		ixs = field.lr1[0]-1
		iys = field.lr1[1]-1
		izs = max(field.lr1[2]-1,1)
	else:
		ixs = 1
		iys = 1
		izs = 1
	#
	iix = range(0,field.lr1[0],ixs); nix = len(iix)
	iiy = range(0,field.lr1[1],iys); niy = len(iiy)
	iiz = range(0,field.lr1[2],izs); niz = len(iiz)
	#
	nppel = nix*niy*niz
	nepel = (nix-1)*(niy-1)*max((niz-1),1)
	nel   = field.nel*nepel
	#
	if (field.ndim==3):
		nvert = 8
		cellType = tvtk.Hexahedron().cell_type
	else:
		nvert = 4
		cellType = tvtk.Quad().cell_type
	#
	ct = np.array(nel*[cellType])
	of = np.arange(0,nvert*nel,nvert)
	ce = np.zeros(nel*(nvert+1)); ce[range(0,nel*(nvert+1),nvert+1)] = nvert
	if (field.var[0]!=0):
		r  = np.zeros((nvert*nel,3))
	if (field.var[1]!=0):
		v  = np.zeros((nvert*nel,3))
	if (field.var[2]==1):
		p  = np.zeros((nvert*nel))
	if (field.var[3]==1):
		T  = np.zeros((nvert*nel))
	if (field.var[4]!=0):
		S  = np.zeros((nvert*nel,field.var[4]))
	#
	ice = -(nvert + 1)
	for iel in range(field.nel):
		for iz in range(niz):
			for iy in range(niy):
				for ix in range(nix):
					if (field.var[0]==3):
						r[iel*nppel + ix + iy*nix + iz*nix*niy,:] = field.elem[iel].pos [:, iiz[iz], iiy[iy], iix[ix]]
					if (field.var[1]==3):
						v[iel*nppel + ix + iy*nix + iz*nix*niy,:] = field.elem[iel].vel [:, iiz[iz], iiy[iy], iix[ix]]
					if (field.var[2]==1):
						p[iel*nppel + ix + iy*nix + iz*nix*niy]   = field.elem[iel].pres[:, iiz[iz], iiy[iy], iix[ix]]
					if (field.var[3]==1):
						T[iel*nppel + ix + iy*nix + iz*nix*niy]   = field.elem[iel].temp[:, iiz[iz], iiy[iy], iix[ix]]
					if (field.var[4]!=0):
						S[iel*nppel + ix + iy*nix + iz*nix*niy,:] = field.elem[iel].scal[:, iiz[iz], iiy[iy], iix[ix]]
		if (field.var[0]==3):
			for iz in max(range(niz-1), [0]):
				for iy in range(niy-1):
					for ix in range(nix-1):
						ice = ice + nvert + 1
						for face in range(field.ndim-1):
							ce[ice + face*4 + 1] = iel*nppel + ix   +  iy   *nix + (iz+face)*nix*niy
							ce[ice + face*4 + 2] = iel*nppel + ix+1 +  iy   *nix + (iz+face)*nix*niy
							ce[ice + face*4 + 3] = iel*nppel + ix+1 + (iy+1)*nix + (iz+face)*nix*niy
							ce[ice + face*4 + 4] = iel*nppel + ix   + (iy+1)*nix + (iz+face)*nix*niy

	# create the array of cells
	ca = tvtk.CellArray()
	ca.set_cells(nel, ce)
	# create the unstructured dataset
	dataset = tvtk.UnstructuredGrid(points=r)
	# set the cell types
	dataset.set_cells(ct, of, ca)
	# set the data
	dataset.point_data.vectors = v
	dataset.point_data.vectors.name = 'vel'
	if (field.var[2]==1):
		dataset.point_data.scalars = p
		dataset.point_data.scalars.name = 'pres'
	if (field.var[3]==1):
		dataset.point_data.add_array(T)
		dataset.point_data.get_array(2).name = 'temp'
	if (field.var[4]!=0):
		for ii in range(field.var[4]):
			dataset.point_data.add_array(S[:,ii])
			dataset.point_data.get_array(ii+3).name = 'scal_%d' % (ii+1)
	#
	dataset.point_data.update()
	#
	return dataset



#=============================================================================#
# exadata                                                                     #
#                                                                             #
#                                                                             #
#                                                                             #
# Authors: Jacopo Canton, Nicolo' Fabbiane, Guillaume Chauvat                 #
# Contacts: jacopo.canton(at)gmail.com                                        #
# Last edit: 2020-02-20                                                       #
#=============================================================================#
from functools import reduce
import numpy as np
import copy
from pymech.log import logger


#==============================================================================
class datalims:
	"""A class containing the extrema of all quantities stored in the mesh"""

	def __init__(self, nb_var, elements):
		#                    x,y,z   min,max
		self.pos  = np.zeros((3     , 2))
		#                    u,v,w   min,max
		self.vel  = np.zeros((3     , 2))
		#                    p       min,max
		self.pres = np.zeros((nb_var[2], 2))
		#                    T       min,max
		self.temp = np.zeros((nb_var[3], 2))
		#                    s_i     min,max
		self.scal = np.zeros((nb_var[4], 2))
		#
		self._variables = ("pos", "vel", "pres", "temp", "scal")

		aggregated_lims = reduce(self._lims_aggregator, elements)
		for var in self._variables:
			agg_lims_var = aggregated_lims[var]
			# set minimum
			getattr(self, var)[:, 0] = agg_lims_var[0]
			# set maximum
			getattr(self, var)[:, 1] = agg_lims_var[1]

	def _lims_per_element(self, elem):
		"""Get local limits for a given element."""
		if isinstance(elem, dict):
			return elem

		axis = (1, 2, 3)
		elem_lims = {
			var: (
				getattr(elem, var).min(axis),
				getattr(elem, var).max(axis)
			)
			for var in self._variables
		}
		return elem_lims

	def _lims_aggregator(self, elem1, elem2):
		"""Reduce local limits to global limits."""
		l1 = self._lims_per_element(elem1)
		l2 = self._lims_per_element(elem2)

		aggregated_lims = {
			var: (
				np.minimum(l1[var][0], l2[var][0]),
				np.maximum(l1[var][1], l2[var][1])
			)
			for var in self._variables
		}
		return aggregated_lims


#==============================================================================
class elem:
	"""A class containing one nek element/SIMSON flow field"""

	def __init__(self, var, lr1, nbc):
		#                    x,y,z   lz      ly      lx
		self.pos  = np.zeros((3     , lr1[2], lr1[1], lr1[0]))
		#                    one per edge
		self.curv = np.zeros((12, 5))
		#             curvature type
		self.ccurv = ['' for i in range(12)]
		#                    u,v,w   lz      ly      lx
		self.vel  = np.zeros((3     , lr1[2], lr1[1], lr1[0]))
		#                    p       lz      ly      lx
		self.pres = np.zeros((var[2], lr1[2], lr1[1], lr1[0]))
		#                    T       lz      ly      lx
		self.temp = np.zeros((var[3], lr1[2], lr1[1], lr1[0]))
		#                    s_i     lz      ly      lx
		self.scal = np.zeros((var[4], lr1[2], lr1[1], lr1[0]))
		#                    list of 8 parameters, one per face
		#                    one column for velocity, one for temperature, and one for each scalar
		self.bcs  = np.zeros((nbc, 6), dtype='U3, i4, i4, f8, f8, f8, f8, f8')


	def face_center(self, i):
		"""Return the coordinates (x, y, z) of the center of the face number i"""

		if i == 0:
			kx1, ky1, kz1 =  0,  0,  0
			kx2, ky2, kz2 =  0,  0, -1
			kx3, ky3, kz3 = -1,  0,  0
			kx4, ky4, kz4 = -1,  0, -1
		elif i == 1:
			kx1, ky1, kz1 = -1,  0,  0
			kx2, ky2, kz2 = -1,  0, -1
			kx3, ky3, kz3 = -1, -1,  0
			kx4, ky4, kz4 = -1, -1, -1
		elif i == 2:
			kx1, ky1, kz1 =  0, -1,  0
			kx2, ky2, kz2 =  0, -1, -1
			kx3, ky3, kz3 = -1, -1,  0
			kx4, ky4, kz4 = -1, -1, -1
		elif i == 3:
			kx1, ky1, kz1 =  0,  0,  0
			kx2, ky2, kz2 =  0,  0, -1
			kx3, ky3, kz3 =  0, -1,  0
			kx4, ky4, kz4 =  0, -1, -1
		elif i == 4:
			kx1, ky1, kz1 =  0,  0,  0
			kx2, ky2, kz2 =  0, -1,  0
			kx3, ky3, kz3 = -1,  0,  0
			kx4, ky4, kz4 = -1, -1,  0
		elif i == 5:
			kx1, ky1, kz1 =  0,  0, -1
			kx2, ky2, kz2 =  0, -1, -1
			kx3, ky3, kz3 = -1,  0, -1
			kx4, ky4, kz4 = -1, -1, -1
		else:
			logger.error('Invalid face number {} (must be between 0 and 5)'.format(i))
		(x1, y1, z1) = self.pos[:, kz1, ky1, kx1]
		(x2, y2, z2) = self.pos[:, kz2, ky2, kx2]
		(x3, y3, z3) = self.pos[:, kz3, ky3, kx3]
		(x4, y4, z4) = self.pos[:, kz4, ky4, kx4]
		return (0.25*(x1+x2+x3+x4), 0.25*(y1+y2+y3+y4), 0.25*(z1+z2+z3+z4))


#==============================================================================
class exadata:
	"""A class containing data for reading/writing binary simulation files"""

	def __init__(self, ndim, nel, lr1, var, nbc=0):
		self.ndim   = ndim
		self.nel    = nel
		self.ncurv  = []
		self.nbc    = nbc
		self.var    = var
		self.lr1    = lr1
		self.time   = []
		self.istep  = []
		self.wdsz   = []
		self.endian = []
		self.elem   = [elem(var, lr1, nbc) for i in range(nel)]
		self.elmap  = np.linspace(1, nel, nel, dtype=np.int32)

	@property
	def lims(self):
		return datalims(self.var, self.elem)

	def check_connectivity(self):
		"""Check element connectivity, specifically for matching boundary
		conditions and geometry. Errors are reported as logging messages.

		"""
		dim = self.ndim
		err = False
		for (iel, el) in enumerate(self.elem):
			for ibc in range(self.nbc):
				for iface in range(2*dim):
					cbc = el.bcs[ibc, iface][0]
					if cbc == 'E' or cbc == 'P':
						connected_iel = int(el.bcs[ibc, iface][3])-1
						connected_face = int(el.bcs[ibc, iface][4])-1
						cbc1 = self.elem[connected_iel].bcs[ibc, connected_face][0]
						iel1 = int(self.elem[connected_iel].bcs[ibc, connected_face][3])-1
						iface1 = int(self.elem[connected_iel].bcs[ibc, connected_face][4])-1
						if iel1 < 0 or iel1 >= self.nel:
							err = True
							logger.error(
								"face {:} of element {:} is connected to face {:} of the nonexistent element {:}"
								.format(iface, iel, connected_face, connected_iel)
							)
						else:
							if cbc1 != cbc:
								err = True
								logger.error(
									"mismatched boundary conditions: face {:} of element {:} with condition {:} is connected to face {:} of element {:} with condition {:}"
									.format(iface+1, iel+1, cbc, connected_face+1, connected_iel+1, cbc1)
								)
							if iel1 != iel:
								err = True
								logger.error(
									"mismatched elements: face {:} of element {:} is connected to face {:} of element {:} but that face is connected to face {:} of element {:}"
									.format(iface+1, iel+1, connected_face+1, connected_iel+1, iface1+1, iel1+1)
								)
							if iface1 != iface:
								err = True
								logger.error(
									"mismatched faces: face {:} of element {:} is connected to face {:} of element {:} but that face is connected to face {:} of element {:}"
									.format(iface+1, iel+1, connected_face+1, connected_iel+1, iface1+1, iel1+1)
								)
		return not err

	def merge(self, other, tol=1e-9, ignore_empty=True):
		"""
		Merges another exadata into the current one and connects it

		Parameters
		----------
		other: exadata
			mesh to merge into self

		tol: float
			maximum distance at which points are considered identical

		ignore_empty: bool
			if True, the faces with an empty boundary condition ('') will be treated as internal faces and will not be merged.
			This is useful if internal boundary conditions are not defined and will make the operation *much* faster,
			but requires boundary conditions to be defined on the faces to be merged.

		"""

		# perform some consistency checks
		if self.ndim != other.ndim:
			logger.error('Cannot merge meshes of dimensions {} and {}!'.format(self.ndim, other.ndim))
			return -1
		if self.lr1[0] != other.lr1[0]:
			logger.error('Cannot merge meshes of different polynomial orders ({} != {})'.format(self.lr1[0], other.lr1[0]))
			return -2

		# add the new elements (in an inconsistent state if there are internal boundary conditions)
		nel1 = self.nel
		self.nel = self.nel + other.nel
		self.ncurv = self.ncurv + other.ncurv
		self.elmap = np.concatenate((self.elmap, other.elmap+nel1))
		# the deep copy is required here to avoid leaving the 'other' mesh in an inconsistent state by modifying its elements
		self.elem = self.elem + copy.deepcopy(other.elem)

		# check how many boundary condition fields we have
		nbc = min(self.nbc, other.nbc)

		# correct the boundary condition numbers:
		# the index of the neighbours have changed
		for iel in range(nel1, self.nel):
			for ibc in range(other.nbc):
				for iface in range(6):
					bc = self.elem[iel].bcs[ibc, iface][0]
					if bc == 'E' or bc == 'P':
						neighbour = self.elem[iel].bcs[ibc, iface][3]
						self.elem[iel].bcs[ibc, iface][3] = neighbour + nel1

		# glue common faces together
		# only look for the neighbour in the first BC field because it should be the same in all fields.
		# FIXME: this will fail to correct periodic conditions if periodic domains are merged together.
		nfaces = 2*self.ndim
		nchanges = 0  # counter for the boundary conditions connected
		if nbc > 0:
			for iel in range(nel1, self.nel):
				for iface in range(nfaces):
					bc = self.elem[iel].bcs[0, iface][0]
					if bc != 'E' and not (ignore_empty and bc == ''):
						# boundary element, look if it can be connected to something
						for iel1 in range(nel1):
							for iface1 in range(nfaces):
								bc1 = self.elem[iel1].bcs[0, iface1][0]
								if bc1 != 'E' and not (ignore_empty and bc1 == ''):
									# if the centers of the faces are close, connect them together
									x0, y0, z0 =  self.elem[iel ].face_center(iface)
									x1, y1, z1 =  self.elem[iel1].face_center(iface1)
									dist2 = (x1-x0)**2 + (y1-y0)**2 + (z1-z0)**2
									if dist2 <= tol**2:
										# reconnect the periodic faces together (assumes that all fields are periodic)
										if bc == 'P' and bc1 == 'P':
											iel_p    = int(self.elem[iel ].bcs[0, iface ][3])-1
											iel_p1   = int(self.elem[iel1].bcs[0, iface1][3])-1
											iface_p  = int(self.elem[iel ].bcs[0, iface ][4])-1
											iface_p1 = int(self.elem[iel1].bcs[0, iface1][4])-1
											for ibc in range(nbc):
												self.elem[iel_p ].bcs[ibc, iface_p ][0] = 'P'
												self.elem[iel_p1].bcs[ibc, iface_p1][0] = 'P'
												self.elem[iel_p ].bcs[ibc, iface_p ][3] = iel_p1+1
												self.elem[iel_p1].bcs[ibc, iface_p1][3] = iel_p +1
												self.elem[iel_p ].bcs[ibc, iface_p ][4] = iface_p1+1
												self.elem[iel_p1].bcs[ibc, iface_p1][4] = iface_p +1
										for ibc in range(nbc):
											self.elem[iel ].bcs[ibc, iface ][0] = 'E'
											self.elem[iel1].bcs[ibc, iface1][0] = 'E'
											self.elem[iel ].bcs[ibc, iface ][3] = iel1+1
											self.elem[iel1].bcs[ibc, iface1][3] = iel +1
											self.elem[iel ].bcs[ibc, iface ][4] = iface1+1
											self.elem[iel1].bcs[ibc, iface1][4] = iface +1
										nchanges = nchanges+1
		logger.debug('merged {} pairs of faces'.format(nchanges))
		return nchanges



#=============================================================================#
# simsonsuite                                                                 #
#                                                                             #
# A python module for reading and writing SIMSON files                        #
#                                                                             #
# Authors: Jacopo Canton, Nicolo' Fabbiane                                    #
# Contacts: jcanton(at)mech.kth.se, nicolo(at)mech.kth.se                     #
# Last edit: 2020-02-19                                                       #
#=============================================================================#
import struct
import numpy as np
import pymech.exadata as exdat

#==============================================================================
def readdns(fname):
	"""A function for reading binary data from the SIMSON binary format

	Parameters
	----------
	fname : str
		file name
	"""
	#
	try:
		infile = open(fname, 'rb')
	except IOError as e:
		print('I/O error ({0}): {1}'.format(e.errno, e.strerror))
		return -1
	#
	#---------------------------------------------------------------------------
	# READ HEADER
	#---------------------------------------------------------------------------
	wdsz = 8
	realtype = 'd'
	#
	# identify endian encoding (and number passive scalar)
	etagb = infile.read(4)
	etagL = struct.unpack('<i', etagb)[0]
	etagB = struct.unpack('>i', etagb)[0]
	# 1644 = 44 + (2*8) * 100 means a maximum of 100 passive scalars
	if ((etagL >= 44) & (etagL <= 1644)):
		# print('Reading little-endian file\n')
		emode = '<'
		nscal = int((etagL-44)/(2*wdsz))
	elif ((etagB >= 44) & (etagB <= 1644)):
		# print('Reading big-endian file\n')
		emode = '>'
		nscal = int((etagB-44)/(2*wdsz))
	else:
		print('ERROR: could not initerpret endianness')
		return -3
	#
	# Reynolds Number
	Re = infile.read(wdsz)
	Re = struct.unpack(emode+realtype,Re)[0]
	#
	# Poiseuille/Couette flag [deprecated]
	PouCou = infile.read(4)
	PouCou = struct.unpack(emode+'i',PouCou)[0]
	#
	# physical box size (here x and z only)
	boxsz = [0 for i in range(3)]
	dum = infile.read(2*wdsz)
	dum = list(struct.unpack(emode+2*realtype,dum))
	boxsz[0] = dum[0]; boxsz[2] = dum[1]
	#
	# time
	time = infile.read(wdsz)
	time = struct.unpack(emode+realtype,time)[0]
	#
	# dummy variable
	dum = infile.read(wdsz)
	#
	# passive scalar(s) (not tested)
	pr = np.zeros(nscal)
	m  = np.zeros(nscal)
	for i in range(nscal):
		pr[i] = infile.read(wdsz)
		pr[i] = struct.unpack(emode+realtype,pr[i])[0]
		m[i] = infile.read(wdsz)
		m[i] = struct.unpack(emode+realtype,m[i])[0]
	#
	# end-of-line
	eol = infile.read(8)
	#
	# box size
	lr1 = infile.read(3*4)
	lr1 = list(struct.unpack(emode+3*'i',lr1))
	#
	# nfzsym (z-symmetry flag)
	nfzsym = infile.read(4)
	nfzsym = list(struct.unpack(emode+'i',nfzsym))[0]
	#
	# end-of-line
	eol = infile.read(8)
	#
	# compute total number of points per element
	npel = lr1[0] * lr1[1] * lr1[2]
	#
	# get number of pysical dimensions
	ndim = 2 + (lr1[2]>1)
	#
	# flow type
	fltype = infile.read(4)
	fltype = struct.unpack(emode+'i',fltype)[0]
	#
	# delta star (and boxsz along y-direction)
	dstar = infile.read(wdsz)
	dstar = struct.unpack(emode+realtype,dstar)[0]
	boxsz[1] = 2/dstar
	#
	# end-of-line
	eol = infile.read(8)
	#
	# flow-type dependent quantities
	#
	if (fltype == -1):
		rlam = infile.read(wdsz)
		rlam = struct.unpack(emode+realtype,rlam)[0]
		eol = infile.read(8)
	if (fltype == -2):
		rlam = infile.read(wdsz)
		rlam = struct.unpack(emode+realtype,rlam)[0]
		spanv = infile.read(wdsz)
		spanv = struct.unpack(emode+realtype,spanv)[0]
		eol = infile.read(8)
	if ((fltype == 4) or (fltype == 5)):
		bstart = infile.read(wdsz)
		bstart = struct.unpack(emode+realtype,bstart)[0]
		blength = infile.read(wdsz)
		blength = struct.unpack(emode+realtype,blength)[0]
		eol = infile.read(8)
	if ((fltype >=4) and (fltype <= 9)):
		bstart = infile.read(wdsz)
		bstart = struct.unpack(emode+realtype,bstart)[0]
		blength = infile.read(wdsz)
		blength = struct.unpack(emode+realtype,blength)[0]
		rlam = infile.read(wdsz)
		rlam = struct.unpack(emode+realtype,rlam)[0]
		spanv = infile.read(wdsz)
		spanv = struct.unpack(emode+realtype,spanv)[0]
		eol = infile.read(8)
	if (abs(fltype) == 20):
		gr = infile.read(nscal*wdsz)
		gr = struct.unpack(emode+nscal*realtype, gr)
		eol = infile.read(8)
	#
	# get variables
	var = [0 for i in range(5)]
	var[0] = ndim  # position
	var[1] = ndim  # velocity
	var[2] = 0     # pressure is not saved (SIMSON)
	var[3] = 0     # temperature is treated like a scalar (SIMSON)
	var[4] = nscal # scalars
	#
	#---------------------------------------------------------------------------
	# READ DATA
	#---------------------------------------------------------------------------
	#
	# number of points
	npel = lr1[0]*lr1[1]*lr1[2]
	#
	# number of points per plane
	nppl = lr1[0]*lr1[2]
	#
	# reading buffer in fourier space
	fou =    np.zeros((lr1[2],lr1[1],lr1[0]//2+1)) + \
			1j*np.zeros((lr1[2],lr1[1],lr1[0]//2+1))
	#
	# initialize data structure
	data = exdat.exadata(ndim, 1, lr1, var)
	data.time   = time
	data.wdsz   = wdsz
	if (emode == '<'):
		data.endian = 'little'
	elif (emode == '>'):
		data.endian = 'big'
	#
	# generate geometry
	# - x-direction
	dx = boxsz[0]/lr1[0]
	for ix in range(lr1[0]):
		data.elem[0].pos[0,:,:,ix] = dx * ix
	# - y-direction
	dy = np.arccos(-1.0)/(lr1[1]-1)
	for iy in range(lr1[1]):
		data.elem[0].pos[1,:,iy,:] = boxsz[1] * (1-np.cos(dy*iy))/2
	# - z-direction
	dz = boxsz[2]/lr1[2]
	for iz in range(lr1[2]):
		data.elem[0].pos[2,iz,:,:] = dz * (iz-lr1[2]/2)
	#
	# read velocity and transform in physical space
	for idim in range(3):
		for iz in range(lr1[2]):
			if (iz <= lr1[2]/2):
				izf = iz
			else:
				izf = lr1[2]//2 * 3 - (iz+1)
			for iy in range(lr1[1]):
				fi = infile.read(lr1[0]*wdsz)
				fi = list(struct.unpack(emode+lr1[0]*realtype, fi))
				ip = 0
				for ix in range(int(lr1[0]//2)):
					fou[izf,iy,ix] = (fi[ip] + 1j*fi[ip+1]) * nppl * (-1)**idim
					ip += 2
				# end-of-line
				eol = infile.read(8)
		#
		# back to physical space
		data.elem[0].vel[idim,:,:,:] = np.fft.irfft2(fou,(lr1[0],lr1[2]),(2,0))
	#
	# read scalars and transform in physical space
	for ivar in range(var[4]):
		for iz in range(lr1[2]):
			if (iz <= lr1[2]/2):
				izf = iz
			else:
				izf = lr1[2]//2 * 3 - (iz+1)
			for iy in range(lr1[1]):
				fi = infile.read(lr1[0]*wdsz)
				fi = list(struct.unpack(emode+lr1[0]*realtype, fi))
				ip = 0
				for ix in range(int(lr1[0]//2)):
					fou[izf,iy,ix] = (fi[ip] + 1j*fi[ip+1]) * nppl
					ip += 2
				# end-of-line
				eol = infile.read(8)
		#
		# back to physical space
		data.elem[0].scal[ivar,:,:,:] = np.fft.irfft2(fou,(lr1[0],lr1[2]),(2,0))
	#
	#---------------------------------------------------------------------------
	# CLOSE FILE
	#---------------------------------------------------------------------------
	#
	# close file
	infile.close()
	#
	# output
	return data


#==============================================================================
def readplane(fname):
	"""A function for reading binary data from SIMSON's (pxyst) plane files

	Parameters
	----------
	fname : str
		file name
	"""
	#
	try:
		infile = open(fname, 'rb')
	except IOError as e:
		print('I/O error ({0}): {1}'.format(e.errno, e.strerror))
		return -1
	#
	#---------------------------------------------------------------------------
	# READ HEADER
	#---------------------------------------------------------------------------
	wdsz = 8
	realtype = 'd'
	#
	# identify endian encoding (and number passive scalar)
	etagb = infile.read(4)
	etagL = struct.unpack('<i', etagb)[0]
	etagB = struct.unpack('>i', etagb)[0]
	# etagL=1000 for some reason (see rrr.m)
	if (etagL <= 1000):
		# print('Reading little-endian file\n')
		endian = 'little'
		emode = '<'
		ndim = int(etagL/4)
	else:
		# print('Reading big-endian file\n')
		endian = 'big'
		emode = '>'
		ndim = int(etagB/4)
	#
	nt=1
	#nn = np.zeros(ndim)
	nn = []
	for i in range(ndim):
		nnn = infile.read(4)
		nn.append( int(struct.unpack(emode+'i',nnn)[0]) )
		nt = nt*nn[i]
	#
	# end-of-line
	eol = infile.read(4)
	#
	if (ndim==1):
		print('reading file %s (%s endian): %d' %(fname,endian,nn[0]))
		x = np.zeros(nn[0])
	elif (ndim==2):
		print('reading file %s (%s endian): %d x %d' %(fname,endian,nn[0],nn[1]))
		x = np.zeros((2,nn[0],nn[1]))
	elif (ndim==3):
		print('reading file %s (%s endian): %d x %d x %d' %(fname,endian,nn[0],nn[1],nn[2]))
		x = np.zeros((3,nn[0],nn[1],nn[2]))
	else:
		print('ERROR: more than three dimensions')
		return -3
	#
	#---------------------------------------------------------------------------
	# READ COORDINATES
	#---------------------------------------------------------------------------
	if (ndim==1):
		# TODO untested
		print('WARNING: reading 1D files was not tested')
		eol = infile.read(4)
		nt1 = struct.unpack(emode+'i',eol)[0]/8
		dum = infile.read(nt1*wdsz)
		x   = struct.unpack(emode+nt1*realtype,dum)
		#
		# end-of-line
		eol = infile.read(4)
	elif (ndim==2):
		for i in range(ndim):
			eol = infile.read(4)
			dum = infile.read(nt*wdsz)
			xx  = struct.unpack(emode+nt*realtype,dum)
			x[i,:,:] = np.reshape(xx, (nn[0],nn[1]),'F')
			#
			# end-of-line
			eol = infile.read(4)
	elif (ndim==3):
		# TODO untested
		print('WARNING: reading 3D files was not tested')
		for i in range(ndim):
			eol = infile.read(4)
			nt3 = struct.unpack(emode+'i',eol)[0]/8
			dum = infile.read(nt3*wdsz)
			xx  = struct.unpack(emode+nt3*realtype,dum)
			x[i,:,:,:] = np.reshape(xx, (nn[0],nn[1],nn[2]),'F')
			#
			# end-of-line
			eol = infile.read(4)
	#
	#---------------------------------------------------------------------------
	# READ DATA
	#---------------------------------------------------------------------------
	eol = infile.read(4)
	dum = infile.read(nt*wdsz)
	dd  = struct.unpack(emode+nt*realtype,dum)
	if (ndim==1):
		# TODO untested
		d = np.reshape(dd, (nn[0]),'F')
	elif (ndim==2):
		d = np.reshape(dd, (nn[0],nn[1]),'F')
	elif (ndim==3):
		# TODO untested
		d = np.reshape(dd, (nn[0],nn[1],nn[2]),'F')
	#
	#---------------------------------------------------------------------------
	# CLOSE FILE
	#---------------------------------------------------------------------------
	#
	# close file
	infile.close()
	#
	# output
	return x, d, nn, ndim



"""Interface for reading files as xarray_ datasets.

.. _xarray: https://xarray.pydata.org

"""
from pathlib import Path
from glob import glob

import numpy as np
import xarray as xr

from .neksuite import readnek


def open_dataset(path, **kwargs):
	"""Helper function for opening a file as an xarray_ dataset.

	Parameters
	----------
	path : str
		Path to a field file (only Nek files are supported at the moment.)

	kwargs : dict
		Keyword arguments passed on to the compatible open function.

	"""
	path = Path(path)
	if path.suffix.startswith('.f'):
		_open = _open_nek_dataset
	else:
		raise NotImplementedError(
			"Filetype: {} is not supported.".format(path.suffix)
		)

	return _open(path, **kwargs)


def open_mfdataset(
	paths,
	chunks=None,
	concat_dim="time",
	compat="no_conflicts",
	preprocess=None,
	engine=None,
	lock=None,
	data_vars="all",
	coords="different",
	combine="nested",
	autoclose=None,
	parallel=False,
	join="outer",
	attrs_file=None,
	**kwargs,
):
	"""Helper function for opening multiple files as an xarray_ dataset.
	Adapted from upstream implementation_. See docs_ for usage.

	.. todo::

		To be removed when a backend entrypoint_ is implementated.

	.. _implementation: https://github.com/pydata/xarray/blob/484d1ce5ff8969b6ca6fa942b344379725f33b9c/xarray/backends/api.py#L726
	.. _docs: https://xarray.pydata.org/en/v0.15.1/generated/xarray.open_mfdataset.html
	.. _entrypoint: https://github.com/pydata/xarray/pull/3166

	"""
	if isinstance(paths, str):
		paths = sorted(glob(paths))
	else:
		paths = [str(p) if isinstance(p, Path) else p for p in paths]

	if not paths:
		raise OSError("no files to open")

	# If combine='by_coords' then this is unnecessary, but quick.
	# If combine='nested' then this creates a flat list which is easier to
	# iterate over, while saving the originally-supplied structure as "ids"
	if combine == "nested":
		if isinstance(concat_dim, (str, xr.DataArray)) or concat_dim is None:
			concat_dim = [concat_dim]

	open_kwargs = dict()

	if parallel:
		import dask

		# wrap the open_dataset, getattr, and preprocess with delayed
		open_ = dask.delayed(open_dataset)
		if preprocess is not None:
			preprocess = dask.delayed(preprocess)
	else:
		open_ = open_dataset

	datasets = [open_(p, **open_kwargs) for p in paths]
	if preprocess is not None:
		datasets = [preprocess(ds) for ds in datasets]

	if parallel:
		# calling compute here will return the datasets
		# the underlying datasets will still be stored as dask arrays
		datasets, = dask.compute(datasets)

	# Combine all datasets, closing them in case of a ValueError
	try:
		if combine == "nested":
			# Combined nested list by successive concat and merge operations
			# along each dimension, using structure given by "ids"
			combined = xr.combine_nested(
				datasets,
				concat_dim=concat_dim,
				compat=compat,
				data_vars=data_vars,
				coords=coords,
				join=join,
			)
		elif combine == "by_coords":
			# Redo ordering from coordinates, ignoring how they were ordered
			# previously
			combined = xr.combine_by_coords(
				datasets, compat=compat, data_vars=data_vars, coords=coords, join=join
			)
		else:
			raise ValueError(
				"{} is an invalid option for the keyword argument"
				" ``combine``".format(combine)
			)
	except ValueError:
		for ds in datasets:
			ds.close()
		raise

	# read global attributes from the attrs_file or from the first dataset
	if attrs_file is not None:
		if isinstance(attrs_file, Path):
			attrs_file = str(attrs_file)
		combined.attrs = datasets[paths.index(attrs_file)].attrs
	else:
		combined.attrs = datasets[0].attrs

	return combined


def _open_nek_dataset(path):
	"""Interface for converting Nek field files into xarray_ datasets."""
	field = readnek(path)
	if isinstance(field, int):
		raise IOError("Failed to load {}".format(path))

	elements = field.elem
	elem_stores = [_NekDataStore(elem) for elem in elements]
	elem_dsets = [
		xr.Dataset.load_store(store).set_coords(store.axes)
		for store in elem_stores
	]

	# See: https://github.com/MITgcm/xmitgcm/pull/200
	if xr.__version__ < '0.15.2':
		ds = xr.combine_by_coords(elem_dsets)
	else:
		ds = xr.combine_by_coords(elem_dsets, combine_attrs="drop")

	ds.coords.update({"time": field.time})

	return ds


class _NekDataStore(xr.backends.common.AbstractDataStore):
	"""Xarray store for a Nek field element."""
	def __init__(self, elem):
		self.elem = elem
		self.axes = ("z", "y", "x")

	def meshgrid_to_dim(self, mesh):
		"""Reverse of np.meshgrid. This method extracts one-dimensional
		coordinates from a cubical array format for every direction

		"""
		dim = np.unique(np.round(mesh, 8))
		return dim

	def get_dimensions(self):
		return self.axes

	def get_attrs(self):
		elem = self.elem
		attrs = {
			"boundary_conditions": elem.bcs,
			"curvature": elem.curv,
			"curvature_type": elem.ccurv,
		}
		return attrs

	def get_variables(self):
		"""Generate an xarray dataset from a single element."""
		ax = self.axes
		elem = self.elem

		data_vars = {
			ax[2]: self.meshgrid_to_dim(elem.pos[0]),  # x
			ax[1]: self.meshgrid_to_dim(elem.pos[1]),  # y
			ax[0]: self.meshgrid_to_dim(elem.pos[2]),  # z
			"xmesh": (ax, elem.pos[0]),
			"ymesh": (ax, elem.pos[1]),
			"zmesh": (ax, elem.pos[2]),
			"ux": (ax, elem.vel[0]),
			"uy": (ax, elem.vel[1]),
			"uz": (ax, elem.vel[2]),
		}
		if elem.pres.size:
			data_vars["pressure"] = ax, elem.pres[0]

		if elem.temp.size:
			data_vars["temperature"] = ax, elem.temp[0]

		if elem.scal.size:
			data_vars.update(
				{
					"s{:02d}".format(iscalar + 1): (ax, elem.scal[iscalar])
					for iscalar in range(elem.scal.shape[0])
				}
			)

		return data_vars



#=============================================================================#
# neksuite                                                                    #
#                                                                             #
# A python module for reading and writing nek5000 files                       #
#                                                                             #
# Authors: Jacopo Canton, Nicolo' Fabbiane, Guillaume Chauvat                 #
# Contacts: jacopo.canton(at)gmail.com                                        #
# Last edit: 2020-02-21                                                       #
#=============================================================================#
import struct
import numpy as np
import pymech.exadata as exdat
import sys

from pymech.log import logger


#==============================================================================
def readnek(fname):
	"""A function for reading binary data from the nek5000 binary format

	Parameters
	----------
	fname : str
		file name
	"""
	#
	try:
		infile = open(fname, 'rb')
	except IOError as e:
		logger.critical('I/O error ({0}): {1}'.format(e.errno, e.strerror))
		return -1
	#
	#---------------------------------------------------------------------------
	# READ HEADER
	#---------------------------------------------------------------------------
	#
	# read header
	header = infile.read(132).split()
	logger.debug('Header: {}'.format(b' '.join(header).decode('utf-8')))

	# get word size: single or double precision
	wdsz = int(header[1])
	if (wdsz == 4):
		realtype = 'f'
	elif (wdsz == 8):
		realtype = 'd'
	else:
		logger.error('Could not interpret real type (wdsz = %i)' % (wdsz))
		return -2
	#
	# get polynomial order
	lr1 = [
		int(header[2]),
		int(header[3]),
		int(header[4])
	]
	#
	# compute total number of points per element
	npel = lr1[0] * lr1[1] * lr1[2]
	#
	# get number of physical dimensions
	ndim = 2 + (lr1[2]>1)
	#
	# get number of elements
	nel = int(header[5])
	#
	# get number of elements in the file
	nelf = int(header[6])
	#
	# get current time
	time = float(header[7])
	#
	# get current time step
	istep = int(header[8])
	#
	# get file id
	fid = int(header[9])
	#
	# get tot number of files
	nf = int(header[10])
	#
	# get variables [XUPTS[01-99]]
	variables = header[11].decode('utf-8')
	logger.debug("Variables: {}".format(variables))
	var = [0 for i in range(5)]
	for v in variables:
		if (v == 'X'):
			var[0] = ndim
		elif (v == 'U'):
			var[1] = ndim
		elif (v == 'P'):
			var[2] = 1
		elif (v == 'T'):
			var[3] = 1
		elif (v == 'S'):
			# For example: variables = 'XS44'
			index_s = variables.index('S')
			nb_scalars = int(variables[index_s+1:])
			var[4] = nb_scalars
	#
	# identify endian encoding
	etagb = infile.read(4)
	etagL = struct.unpack('<f', etagb)[0]; etagL = int(etagL*1e5)/1e5
	etagB = struct.unpack('>f', etagb)[0]; etagB = int(etagB*1e5)/1e5
	if (etagL == 6.54321):
		logger.debug('Reading little-endian file\n')
		emode = '<'
	elif (etagB == 6.54321):
		logger.debug('Reading big-endian file\n')
		emode = '>'
	else:
		logger.error('Could not interpret endianness')
		return -3
	#
	# read element map for the file
	elmap = infile.read(4*nelf)
	elmap = list(struct.unpack(emode+nelf*'i', elmap))
	#
	#---------------------------------------------------------------------------
	# READ DATA
	#---------------------------------------------------------------------------
	#
	# initialize data structure
	data = exdat.exadata(ndim, nel, lr1, var, 0)
	data.time   = time
	data.istep  = istep
	data.wdsz   = wdsz
	data.elmap  = np.array(elmap, dtype=np.int32)
	if (emode == '<'):
		data.endian = 'little'
	elif (emode == '>'):
		data.endian = 'big'
	#
	def read_file_into_data(data_var, index_var):
		"""Read binary file into an array attribute of ``data.elem``"""
		fi = infile.read(npel*wdsz)
		fi = np.frombuffer(fi, dtype=emode+realtype, count=npel)

		# Replace elem array in-place with
		# array read from file after reshaping as
		elem_shape = lr1[::-1]  # lz, ly, lx
		data_var[index_var, ...] = fi.reshape(elem_shape)
	#
	# read geometry
	for iel in elmap:
		el = data.elem[iel-1]
		for idim in range(var[0]): # if var[0] == 0, geometry is not read
			read_file_into_data(el.pos, idim)
	#
	# read velocity
	for iel in elmap:
		el = data.elem[iel-1]
		for idim in range(var[1]): # if var[1] == 0, velocity is not read
			read_file_into_data(el.vel, idim)
	#
	# read pressure
	for iel in elmap:
		el = data.elem[iel-1]
		for ivar in range(var[2]): # if var[2] == 0, pressure is not read
			read_file_into_data(el.pres, ivar)
	#
	# read temperature
	for iel in elmap:
		el = data.elem[iel-1]
		for ivar in range(var[3]): # if var[3] == 0, temperature is not read
			read_file_into_data(el.temp, ivar)
	#
	# read scalar fields
	#
	# NOTE: This is not a bug!
	# Unlike other variables, scalars are in the outer loop and elements
	# are in the inner loop
	#
	for ivar in range(var[4]): # if var[4] == 0, scalars are not read
		for iel in elmap:
			el = data.elem[iel-1]
			read_file_into_data(el.scal, ivar)
	#
	#
	# close file
	infile.close()
	#
	# output
	return data


#==============================================================================
def writenek(fname, data):
	"""A function for writing binary data in the nek5000 binary format

	Parameters
	----------
	fname : str
		file name
	data : exadata
		data organised as readnek() output
	"""
	#
	try:
		outfile = open(fname, 'wb')
	except IOError as e:
		logger.critical('I/O error ({0}): {1}'.format(e.errno, e.strerror))
		return -1
	#
	#---------------------------------------------------------------------------
	# WRITE HEADER
	#---------------------------------------------------------------------------
	#
	# multiple files (not implemented)
	fid = 0
	nf = 1
	nelf = data.nel
	#
	# get fields to be written
	vars = ''
	if (data.var[0] > 0): vars += 'X'
	if (data.var[1] > 0): vars += 'U'
	if (data.var[2] > 0): vars += 'P'
	if (data.var[3] > 0): vars += 'T'
	if (data.var[4] > 0):
		# TODO: check if header for scalars are written with zeros filled as S01
		vars += 'S{:02d}'.format(data.var[4])
	#
	# get word size
	if (data.wdsz == 4):
		logger.debug('Writing single-precision file')
	elif (data.wdsz == 8):
		logger.debug('Writing double-precision file')
	else:
		logger.error('Could not interpret real type (wdsz = %i)' % (data.wdsz))
		return -2
	#
	# generate header
	header = '#std %1i %2i %2i %2i %10i %10i %20.13E %9i %6i %6i %s' %(
		data.wdsz, data.lr1[0], data.lr1[1], data.lr1[2], data.nel, nelf,
		data.time, data.istep, fid, nf, vars)
	#
	# write header
	header = header.ljust(132)
	outfile.write(header.encode('utf-8'))
	#
	# decide endianness
	if data.endian in ('big', 'little'):
		byteswap = data.endian != sys.byteorder
		logger.debug(f'Writing {data.endian}-endian file')
	else:
		byteswap = False
		logger.warning(
			f'Unrecognized endianness {data.endian}, '
			f'writing native {sys.byteorder}-endian file'
		)
	#
	def correct_endianness(a):
		''' Return the array with the requested endianness'''
		if byteswap:
			return a.byteswap()
		else:
			return a
	#
	# write tag (to specify endianness)
	endianbytes = np.array([6.54321], dtype=np.float32)
	correct_endianness(endianbytes).tofile(outfile)
	#
	# write element map for the file
	correct_endianness(data.elmap).tofile(outfile)
	#
	#---------------------------------------------------------------------------
	# WRITE DATA
	#---------------------------------------------------------------------------
	#
	# compute total number of points per element
	npel = data.lr1[0] * data.lr1[1] * data.lr1[2]
	#
	def write_ndarray_to_file(a):
		'''Write a data array to the output file in the requested precision and endianness'''
		if data.wdsz == 4:
			correct_endianness(a.astype(np.float32)).tofile(outfile)
		else:
			correct_endianness(a).tofile(outfile)
	#
	# write geometry
	for iel in data.elmap:
		for idim in range(data.var[0]): # if var[0] == 0, geometry is not written
			write_ndarray_to_file(data.elem[iel-1].pos[idim, :, :, :])
	#
	# write velocity
	for iel in data.elmap:
		for idim in range(data.var[1]): # if var[1] == 0, velocity is not written
			write_ndarray_to_file(data.elem[iel-1].vel[idim, :, :, :])
	#
	# write pressure
	for iel in data.elmap:
		for ivar in range(data.var[2]): # if var[2] == 0, pressure is not written
			write_ndarray_to_file(data.elem[iel-1].pres[ivar, :, :, :])
	#
	# write temperature
	for iel in data.elmap:
		for ivar in range(data.var[3]): # if var[3] == 0, temperature is not written
			write_ndarray_to_file(data.elem[iel-1].temp[ivar, :, :, :])
	#
	# write scalars
	#
	# NOTE: This is not a bug!
	# Unlike other variables, scalars are in the outer loop and elements
	# are in the inner loop
	#
	for ivar in range(data.var[4]): # if var[4] == 0, scalars are not written
		for iel in data.elmap:
			write_ndarray_to_file(data.elem[iel-1].scal[ivar, :, :, :])
	#
	# write max and min of every field in every element (forced to single precision)
	if (data.ndim==3):
		#
		for iel in data.elmap:
			for idim in range(data.var[0]):
				correct_endianness(np.min(data.elem[iel-1].pos[idim, :,:,:]).astype(np.float32)).tofile(outfile)
				correct_endianness(np.max(data.elem[iel-1].pos[idim, :,:,:]).astype(np.float32)).tofile(outfile)
		for iel in data.elmap:
			for idim in range(data.var[1]):
				correct_endianness(np.min(data.elem[iel-1].vel[idim, :,:,:]).astype(np.float32)).tofile(outfile)
				correct_endianness(np.max(data.elem[iel-1].vel[idim, :,:,:]).astype(np.float32)).tofile(outfile)
		for iel in data.elmap:
			for ivar in range(data.var[2]):
				correct_endianness(np.min(data.elem[iel-1].pres[ivar, :,:,:]).astype(np.float32)).tofile(outfile)
				correct_endianness(np.max(data.elem[iel-1].pres[ivar, :,:,:]).astype(np.float32)).tofile(outfile)
		for iel in data.elmap:
			for ivar in range(data.var[3]):
				correct_endianness(np.min(data.elem[iel-1].temp[ivar, :,:,:]).astype(np.float32)).tofile(outfile)
				correct_endianness(np.max(data.elem[iel-1].temp[ivar, :,:,:]).astype(np.float32)).tofile(outfile)
		for iel in data.elmap:
			for ivar in range(data.var[4]):
				correct_endianness(np.min(data.elem[iel-1].scal[ivar, :,:,:]).astype(np.float32)).tofile(outfile)
				correct_endianness(np.max(data.elem[iel-1].scal[ivar, :,:,:]).astype(np.float32)).tofile(outfile)

	# close file
	outfile.close()
	#
	# output
	return 0


#==============================================================================
def readrea(fname):
	"""A function for reading .rea files for nek5000

	Parameters
	----------
	fname : str
		file name
	"""
	#
	try:
		infile = open(fname, 'r')
	except IOError as e:
		logger.critical('I/O error ({0}): {1}'.format(e.errno, e.strerror))
		#return -1
	#
	#---------------------------------------------------------------------------
	# count the number of boundary conditions
	# (it's too dangerous to infer it from the header)
	#---------------------------------------------------------------------------
	#
	nbc = 0
	for line in infile:
		line_split = line.split()
		if 'BOUNDARY' in line_split[2:] and not 'NO' in line_split:
			nbc = nbc + 1

	infile.seek(0)
	#
	#---------------------------------------------------------------------------
	# READ HEADER (2 lines) + ndim + number of parameters
	#---------------------------------------------------------------------------
	#
	infile.readline()
	infile.readline()
	ndim = int(infile.readline().split()[0])
	npar = int(infile.readline().split()[0])
	#
	nface = 2*ndim
	#
	#---------------------------------------------------------------------------
	# READ parameters
	#---------------------------------------------------------------------------
	#
	param = np.zeros((npar,1))
	for ipar in range(npar):
		param[ipar] = float(infile.readline().split()[0])
	#
	#---------------------------------------------------------------------------
	# skip passive scalars
	#---------------------------------------------------------------------------
	#
	npscal_data = int(infile.readline().split()[0])
	for ipscal in range(npscal_data):
		infile.readline()
	#
	#---------------------------------------------------------------------------
	# skip logical switches
	#---------------------------------------------------------------------------
	#
	nswitch = int(infile.readline().split()[0])
	for iswitch in range(nswitch):
		infile.readline()
	#
	#---------------------------------------------------------------------------
	# skip XFAC,YFAC,XZERO,YZERO
	#---------------------------------------------------------------------------
	#
	infile.readline()
	#
	#---------------------------------------------------------------------------
	# READ MESH
	#---------------------------------------------------------------------------
	#
	infile.readline()
	nel = int(infile.readline().split()[0])
	#
	# initialize data structure
	lr1 = [2, 2, ndim-1]
	var = [ndim, 0, 0, 0, 0]
	#
	data = exdat.exadata(ndim, nel, lr1, var, nbc)
	#
	# read geometry
	for iel in range(nel):
		# skip element number and group
		infile.readline()
		for idim in range(var[0]-1): # if ndim == 3 do this twice
			for jdim in range(var[0]):
				fi = infile.readline().split()
				data.elem[iel].pos[jdim,idim,0,0] = float(fi[0])
				data.elem[iel].pos[jdim,idim,0,1] = float(fi[1])
				data.elem[iel].pos[jdim,idim,1,1] = float(fi[2])
				data.elem[iel].pos[jdim,idim,1,0] = float(fi[3])
	#
	#---------------------------------------------------------------------------
	# CURVED SIDE DATA
	#---------------------------------------------------------------------------
	#
	infile.readline()
	ncurved = int(infile.readline().split()[0])
	data.ncurv = ncurved
	for icurved in range(ncurved):
		line = infile.readline()
		if (nel < 1e3):
			iedge = int(line[0:3])-1
			iel   = int(line[3:6])-1
			data.elem[iel].curv[iedge][0] = float(line[6:20])
			data.elem[iel].curv[iedge][1] = float(line[20:34])
			data.elem[iel].curv[iedge][2] = float(line[34:48])
			data.elem[iel].curv[iedge][3] = float(line[48:62])
			data.elem[iel].curv[iedge][4] = float(line[62:76])
			data.elem[iel].ccurv[iedge]   = line[76:79].split()[0]
		elif (nel < 1e6):
			iedge = int(line[0:2])-1
			iel   = int(line[2:8])-1
			data.elem[iel].curv[iedge][0] = float(line[8:22])
			data.elem[iel].curv[iedge][1] = float(line[22:36])
			data.elem[iel].curv[iedge][2] = float(line[36:50])
			data.elem[iel].curv[iedge][3] = float(line[50:64])
			data.elem[iel].curv[iedge][4] = float(line[64:78])
			data.elem[iel].ccurv[iedge]   = line[78:81].split()[0]
		else:
			iedge = int(line[0:2])-1
			iel   = int(line[2:12])-1
			data.elem[iel].curv[iedge][0] = float(line[12:26])
			data.elem[iel].curv[iedge][1] = float(line[26:40])
			data.elem[iel].curv[iedge][2] = float(line[40:54])
			data.elem[iel].curv[iedge][3] = float(line[54:68])
			data.elem[iel].curv[iedge][4] = float(line[68:82])
			data.elem[iel].ccurv[iedge]   = line[82:85].split()[0]
	#
	#---------------------------------------------------------------------------
	# BOUNDARY CONDITIONS
	#---------------------------------------------------------------------------
	#
	infile.readline()  # ***** BOUNDARY CONDITIONS *****
	for ibc in range(nbc):
		infile.readline()  # ***** FLUID   BOUNDARY CONDITIONS ***** [or similar]
		for iel in range(nel):
			for iface in range(nface):
				line = infile.readline()
				if (nel < 1e3):
					data.elem[iel].bcs[ibc, iface][0] = line[1:3].strip()
					data.elem[iel].bcs[ibc, iface][1] = int(line[4:7])
					data.elem[iel].bcs[ibc, iface][2] = int(line[7:10])
					data.elem[iel].bcs[ibc, iface][3] = float(line[10:24])
					data.elem[iel].bcs[ibc, iface][4] = float(line[24:38])
					data.elem[iel].bcs[ibc, iface][5] = float(line[38:52])
					data.elem[iel].bcs[ibc, iface][6] = float(line[52:66])
					data.elem[iel].bcs[ibc, iface][7] = float(line[66:80])
				elif (nel < 1e6):
					data.elem[iel].bcs[ibc, iface][0] = line[1:3].strip()
					data.elem[iel].bcs[ibc, iface][1] = iel + 1
					data.elem[iel].bcs[ibc, iface][2] = iface + 1
					data.elem[iel].bcs[ibc, iface][3] = float(line[10:24])
					data.elem[iel].bcs[ibc, iface][4] = float(line[24:38])
					data.elem[iel].bcs[ibc, iface][5] = float(line[38:52])
					data.elem[iel].bcs[ibc, iface][6] = float(line[52:66])
					data.elem[iel].bcs[ibc, iface][7] = float(line[66:80])
				else:
					data.elem[iel].bcs[ibc, iface][0] = line[1:3].strip()
					data.elem[iel].bcs[ibc, iface][1] = int(line[4:15])
					data.elem[iel].bcs[ibc, iface][2] = int(line[15:16])
					data.elem[iel].bcs[ibc, iface][3] = float(line[16:34])
					data.elem[iel].bcs[ibc, iface][4] = float(line[34:52])
					data.elem[iel].bcs[ibc, iface][5] = float(line[52:70])
					data.elem[iel].bcs[ibc, iface][6] = float(line[70:88])
					data.elem[iel].bcs[ibc, iface][7] = float(line[88:106])
				# ignore some invalid internal 'E' conditions.
				# They are typically written this way by re2torea and Nek5000 ignores them.
				if data.elem[iel].bcs[ibc, iface][0] == 'E' and data.elem[iel].bcs[ibc, iface][3] == 0.:
					data.elem[iel].bcs[ibc, iface][0] = ''
					for j in range(1, 8):
						data.elem[iel].bcs[ibc, iface][j] = 0
		ibc = ibc + 1
	#
	#---------------------------------------------------------------------------
	# FORGET ABOUT WHAT FOLLOWS
	#---------------------------------------------------------------------------
	#
	#
	# close file
	infile.close()
	#
	# output
	return data


#==============================================================================
def writerea(fname, data):
	"""A function for writing ascii .rea files for nek5000

	Parameters
	----------
	fname : str
		file name
	data : exadata
		data organised as in exadata.py
	"""
	#
	try:
		outfile = open(fname, 'w')
	except IOError as e:
		logger.critical('I/O error ({0}): {1}'.format(e.errno, e.strerror))
		#return -1
	#
	#---------------------------------------------------------------------------
	# READ HEADER (2 lines) + ndim + number of parameters
	#---------------------------------------------------------------------------
	#
	outfile.write('****** PARAMETERS ******\n')
	outfile.write('   2.6000     NEKTON VERSION\n')
	outfile.write('   {0:1d} DIMENSIONAL RUN\n'.format(data.ndim))
	outfile.write('         118 PARAMETERS FOLLOW\n')
	outfile.write('   1.00000     P001: DENSITY\n')
	outfile.write('  -1000.00     P002: VISCOSITY\n')
	outfile.write('   0.00000     P003: BETAG\n')
	outfile.write('   0.00000     P004: GTHETA\n')
	outfile.write('   0.00000     P005: PGRADX\n')
	outfile.write('   0.00000     P006: \n')
	outfile.write('   1.00000     P007: RHOCP\n')
	outfile.write('   1.00000     P008: CONDUCT\n')
	outfile.write('   0.00000     P009: \n')
	outfile.write('   0.00000     P010: FINTIME\n')
	outfile.write('   103         P011: NSTEPS\n')
	outfile.write('  -1.00000E-03 P012: DT\n')
	outfile.write('   0.00000     P013: IOCOMM\n')
	outfile.write('   0.00000     P014: IOTIME\n')
	outfile.write('   10          P015: IOSTEP\n')
	outfile.write('   0.00000     P016: PSSOLVER: 0=default\n')
	outfile.write('   1.00000     P017: \n')
	outfile.write('   0.00000     P018: GRID <0 --> # cells on screen\n')
	outfile.write('   0.00000     P019: INTYPE\n')
	outfile.write('   10.0000     P020: NORDER\n')
	outfile.write('   1.00000E-09 P021: DIVERGENCE\n')
	outfile.write('   1.00000E-09 P022: HELMHOLTZ\n')
	outfile.write('   0.00000     P023: NPSCAL\n')
	outfile.write('   1.00000E-02 P024: TOLREL\n')
	outfile.write('   1.00000E-02 P025: TOLABS\n')
	outfile.write('   1.00000     P026: COURANT/NTAU\n')
	outfile.write('   3.00000     P027: TORDER\n')
	outfile.write('   0.00000     P028: TORDER: mesh velocity (0: p28=p27)\n')
	outfile.write('   0.00000     P029: = magnetic visc if > 0, = -1/Rm if < 0\n')
	outfile.write('   0.00000     P030: > 0 ==> properties set in uservp()\n')
	outfile.write('   0.00000     P031: NPERT: #perturbation modes\n')
	outfile.write('   0.00000     P032: #BCs in re2 file, if > 0\n')
	outfile.write('   0.00000     P033: \n')
	outfile.write('   0.00000     P034: \n')
	outfile.write('   0.00000     P035: \n')
	outfile.write('   0.00000     P036: XMAGNET\n')
	outfile.write('   0.00000     P037: NGRIDS\n')
	outfile.write('   0.00000     P038: NORDER2\n')
	outfile.write('   0.00000     P039: NORDER3\n')
	outfile.write('   0.00000     P040: \n')
	outfile.write('   0.00000     P041: 1-->multiplicattive SEMG\n')
	outfile.write('   0.00000     P042: 0=gmres/1=pcg\n')
	outfile.write('   0.00000     P043: 0=semg/1=schwarz\n')
	outfile.write('   0.00000     P044: 0=E-based/1=A-based prec.\n')
	outfile.write('   0.00000     P045: Relaxation factor for DTFS\n')
	outfile.write('   0.00000     P046: reserved\n')
	outfile.write('   0.00000     P047: vnu: mesh material prop.\n')
	outfile.write('   0.00000     P048: \n')
	outfile.write('   0.00000     P049: \n')
	outfile.write('   0.00000     P050: \n')
	outfile.write('   0.00000     P051: \n')
	outfile.write('   0.00000     P052: IOHIS\n')
	outfile.write('   0.00000     P053: \n')
	outfile.write('   0.00000     P054: fixed flow rate dir: |p54|=1,2,3=x,y,z\n')
	outfile.write('   0.00000     P055: vol.flow rate (p54>0) or Ubar (p54<0)\n')
	outfile.write('   0.00000     P056: \n')
	outfile.write('   0.00000     P057: \n')
	outfile.write('   0.00000     P058: \n')
	outfile.write('   0.00000     P059: !=0 --> full Jac. eval. for each el.\n')
	outfile.write('   0.00000     P060: !=0 --> init. velocity to small nonzero\n')
	outfile.write('   0.00000     P061: \n')
	outfile.write('   0.00000     P062: >0 --> force byte_swap for output\n')
	outfile.write('   8.00000     P063: =8 --> force 8-byte output\n')
	outfile.write('   0.00000     P064: =1 --> perturbation restart\n')
	outfile.write('   1.00000     P065: #iofiles (eg, 0 or 64); <0 --> sep. dirs\n')
	outfile.write('   6.00000     P066: output : <0=ascii, else binary\n')
	outfile.write('   6.00000     P067: restart: <0=ascii, else binary\n')
	outfile.write('   0.00000     P068: iastep: freq for avg_all (0=iostep)\n')
	outfile.write('   0.00000     P069:       : freq of srf dump\n')
	outfile.write('   0.00000     P070: \n')
	outfile.write('   0.00000     P071: \n')
	outfile.write('   0.00000     P072: \n')
	outfile.write('   0.00000     P073: \n')
	outfile.write('   0.00000     P074: \n')
	outfile.write('   0.00000     P075: \n')
	outfile.write('   0.00000     P076: \n')
	outfile.write('   0.00000     P077: \n')
	outfile.write('   0.00000     P078: \n')
	outfile.write('   0.00000     P079: \n')
	outfile.write('   0.00000     P080: \n')
	outfile.write('   0.00000     P081: \n')
	outfile.write('   0.00000     P082: \n')
	outfile.write('   0.00000     P083: \n')
	outfile.write('   0.00000     P084: != 0 --> sets initial timestep if p12>0\n')
	outfile.write('   0.00000     P085: dt retio of p84 !=0, for timesteps>0\n')
	outfile.write('   0.00000     P086: reserved\n')
	outfile.write('   0.00000     P087: \n')
	outfile.write('   0.00000     P088: \n')
	outfile.write('   0.00000     P089: \n')
	outfile.write('   0.00000     P090: \n')
	outfile.write('   0.00000     P091: \n')
	outfile.write('   0.00000     P092: \n')
	outfile.write('   20.0000     P093: Number of previous pressure solns saved\n')
	outfile.write('   9.00000     P094: start projecting velocity after p94 step\n')
	outfile.write('   9.00000     P095: start projecting pressure after p95 step\n')
	outfile.write('   0.00000     P096: \n')
	outfile.write('   0.00000     P097: \n')
	outfile.write('   0.00000     P098: \n')
	outfile.write('   3.00000     P099: dealiasing: <0--> off /3--> old /4-->new\n')
	outfile.write('   0.00000     P100: \n')
	outfile.write('   0.00000     P101: Number of additional modes to filter\n')
	outfile.write('   0.00000     P102: Dump out divergence at each time step\n')
	outfile.write('   0.01000     P103: weight of stabilizing filter\n')
	outfile.write('   0.00000     P104: \n')
	outfile.write('   0.00000     P105: \n')
	outfile.write('   0.00000     P106: \n')
	outfile.write('   0.00000     P107: !=0 --> add h2 array in hmholtz eqn\n')
	outfile.write('   0.00000     P108: \n')
	outfile.write('   0.00000     P109: \n')
	outfile.write('   0.00000     P110: \n')
	outfile.write('   0.00000     P111: \n')
	outfile.write('   0.00000     P112: \n')
	outfile.write('   0.00000     P113: \n')
	outfile.write('   0.00000     P114: \n')
	outfile.write('   0.00000     P115: \n')
	outfile.write('   0.00000     P116: \n')
	outfile.write('   0.00000     P117: \n')
	outfile.write('   0.00000     P118: \n')
	outfile.write('      4  Lines of passive scalar data follows2 CONDUCT, 2RHOCP\n')
	outfile.write('   1.00000        1.00000        1.00000        1.00000        1.00000\n')
	outfile.write('   1.00000        1.00000        1.00000        1.00000\n')
	outfile.write('   1.00000        1.00000        1.00000        1.00000        1.00000\n')
	outfile.write('   1.00000        1.00000        1.00000        1.00000\n')
	outfile.write('         13   LOGICAL SWITCHES FOLLOW\n')
	outfile.write(' T      IFFLOW\n')
	outfile.write(' F      IFHEAT\n')
	outfile.write(' T      IFTRAN\n')
	outfile.write(' T F F F F F F F F F F  IFNAV & IFADVC (convection in P.S. fields)\n')
	outfile.write(' F F T T T T T T T T T T  IFTMSH (IF mesh for this field is T mesh)\n')
	outfile.write(' F      IFAXIS\n')
	outfile.write(' F      IFSTRS\n')
	outfile.write(' F      IFSPLIT\n')
	outfile.write(' F      IFMGRID\n')
	outfile.write(' F      IFMODEL\n')
	outfile.write(' F      IFKEPS\n')
	outfile.write(' F      IFMVBD\n')
	outfile.write(' F      IFCHAR\n')
	outfile.write('   2.00000       2.00000      -1.00000      -1.00000     XFAC,YFAC,XZERO,YZERO\n')
	#
	# vertex data
	outfile.write('  ***** MESH DATA *****  6 lines are X,Y,Z;X,Y,Z. Columns corners 1-4;5-8\n')
	outfile.write('  {0:10d} {1:1d} {2:10d} NEL,NDIM,NELV\n'.format(data.nel, data.ndim, data.nel))
	for iel in range(data.nel):
		outfile.write('           ELEMENT {0:10d} [  1a]    GROUP 1\n'.format(iel+1))
		if (data.ndim == 2):
			outfile.write('{0:14.6e}{1:14.6e}{2:14.6e}{3:14.6e}\n'.format(data.elem[iel].pos[0, 0, 0, 0], data.elem[iel].pos[0, 0, 0, -1], data.elem[iel].pos[0, 0, -1, -1], data.elem[iel].pos[0, 0, -1, 0]))
			outfile.write('{0:14.6e}{1:14.6e}{2:14.6e}{3:14.6e}\n'.format(data.elem[iel].pos[1, 0, 0, 0], data.elem[iel].pos[1, 0, 0, -1], data.elem[iel].pos[1, 0, -1, -1], data.elem[iel].pos[1, 0, -1, 0]))
		elif (data.ndim == 3):
			outfile.write('{0:14.6e}{1:14.6e}{2:14.6e}{3:14.6e}\n'.format(data.elem[iel].pos[0, 0, 0, 0], data.elem[iel].pos[0, 0, 0, -1], data.elem[iel].pos[0, 0, -1, -1], data.elem[iel].pos[0, 0, -1, 0]))
			outfile.write('{0:14.6e}{1:14.6e}{2:14.6e}{3:14.6e}\n'.format(data.elem[iel].pos[1, 0, 0, 0], data.elem[iel].pos[1, 0, 0, -1], data.elem[iel].pos[1, 0, -1, -1], data.elem[iel].pos[1, 0, -1, 0]))
			outfile.write('{0:14.6e}{1:14.6e}{2:14.6e}{3:14.6e}\n'.format(data.elem[iel].pos[2, 0, 0, 0], data.elem[iel].pos[2, 0, 0, -1], data.elem[iel].pos[2, 0, -1, -1], data.elem[iel].pos[2, 0, -1, 0]))
			#
			outfile.write('{0:14.6e}{1:14.6e}{2:14.6e}{3:14.6e}\n'.format(data.elem[iel].pos[0, -1, 0, 0], data.elem[iel].pos[0, -1, 0, -1], data.elem[iel].pos[0, -1, -1, -1], data.elem[iel].pos[0, -1, -1, 0]))
			outfile.write('{0:14.6e}{1:14.6e}{2:14.6e}{3:14.6e}\n'.format(data.elem[iel].pos[1, -1, 0, 0], data.elem[iel].pos[1, -1, 0, -1], data.elem[iel].pos[1, -1, -1, -1], data.elem[iel].pos[1, -1, -1, 0]))
			outfile.write('{0:14.6e}{1:14.6e}{2:14.6e}{3:14.6e}\n'.format(data.elem[iel].pos[2, -1, 0, 0], data.elem[iel].pos[2, -1, 0, -1], data.elem[iel].pos[2, -1, -1, -1], data.elem[iel].pos[2, -1, -1, 0]))
	#
	# curved side data
	outfile.write('  ***** CURVED SIDE DATA *****\n')
	outfile.write('  {0:10d} Curved sides follow IEDGE,IEL,CURVE(I),I=1,5, CCURVE\n'.format(data.ncurv))
	for iel in range(data.nel):
		if (data.nel < 1e3):
			#
			for iedge in range(12):
				if (data.elem[iel].ccurv[iedge] != ''):
					outfile.write('{0:3d}{1:3d}{2:14.6e}{3:14.6e}{4:14.6e}{5:14.6e}{6:14.6e}{7:>2s}\n'.format(
						iedge+1, iel+1, data.elem[iel].curv[iedge][0], data.elem[iel].curv[iedge][1], data.elem[iel].curv[iedge][2], data.elem[iel].curv[iedge][3], data.elem[iel].curv[iedge][4], data.elem[iel].ccurv[iedge]))
		elif (data.nel < 1e6):
			#
			for iedge in range(12):
				if (data.elem[iel].ccurv[iedge] != ''):
					outfile.write('{0:2d}{1:6d}{2:14.6e}{3:14.6e}{4:14.6e}{5:14.6e}{6:14.6e}{7:>2s}\n'.format(
						iedge+1, iel+1, data.elem[iel].curv[iedge][0], data.elem[iel].curv[iedge][1], data.elem[iel].curv[iedge][2], data.elem[iel].curv[iedge][3], data.elem[iel].curv[iedge][4], data.elem[iel].ccurv[iedge]))
		else:
			#
			for iedge in range(12):
				if (data.elem[iel].ccurv[iedge] != ''):
					outfile.write('{0:2d}{1:10d}{2:14.6e}{3:14.6e}{4:14.6e}{5:14.6e}{6:14.6e}{7:>2s}\n'.format(
						iedge+1, iel+1, data.elem[iel].curv[iedge][0], data.elem[iel].curv[iedge][1], data.elem[iel].curv[iedge][2], data.elem[iel].curv[iedge][3], data.elem[iel].curv[iedge][4], data.elem[iel].ccurv[iedge]))
	#
	# boundary conditions data
	outfile.write('  ***** BOUNDARY CONDITIONS *****\n')
	outfile.write('  ***** FLUID BOUNDARY CONDITIONS *****\n')
	for ibc in range(data.nbc):
		if ibc == 1:
			outfile.write('  ***** THERMAL BOUNDARY CONDITIONS *****\n')
		elif ibc > 1:
			outfile.write('  ***** PASSIVE SCALAR {:4d} BOUNDARY CONDITIONS *****\n'.format(ibc-1))
		for iel in range(data.nel):
			for iface in range(2*data.ndim):
				# if no boundary condition is specified, write 'E' iel iface [...]
				# this is the behaviour of re2torea.
				if data.elem[iel].bcs[ibc, iface][0] == '':
					data.elem[iel].bcs[ibc, iface][0] = 'E'
					data.elem[iel].bcs[ibc, iface][1] = iel+1
					data.elem[iel].bcs[ibc, iface][2] = iface+1
				if (data.nel < 1e3):
					outfile.write(' {0:2s} {1:3d}{2:3d}{3:14.6e}{4:14.6e}{5:14.6e}{6:14.6e}{7:14.6e}\n'.format(
						data.elem[iel].bcs[ibc, iface][0], data.elem[iel].bcs[ibc, iface][1], data.elem[iel].bcs[ibc, iface][2], data.elem[iel].bcs[ibc, iface][3], data.elem[iel].bcs[ibc, iface][4], data.elem[iel].bcs[ibc, iface][5], data.elem[iel].bcs[ibc, iface][6], data.elem[iel].bcs[ibc, iface][7]))
				elif (data.nel < 1e6):
					outfile.write(' {0:2s} {1:6d}{2:14.6e}{3:14.6e}{4:14.6e}{5:14.6e}{6:14.6e}\n'.format(
						data.elem[iel].bcs[ibc, iface][0], data.elem[iel].bcs[ibc, iface][1], data.elem[iel].bcs[ibc, iface][3], data.elem[iel].bcs[ibc, iface][4], data.elem[iel].bcs[ibc, iface][5], data.elem[iel].bcs[ibc, iface][6], data.elem[iel].bcs[ibc, iface][7]))
				else:
					outfile.write(' {0:2s} {1:11d}{2:1d}{3:18.11e}{4:18.11e}{5:18.11e}{6:18.11e}{7:18.11e}\n'.format(
						data.elem[iel].bcs[ibc, iface][0], data.elem[iel].bcs[ibc, iface][1], data.elem[iel].bcs[ibc, iface][2], data.elem[iel].bcs[ibc, iface][3], data.elem[iel].bcs[ibc, iface][4], data.elem[iel].bcs[ibc, iface][5], data.elem[iel].bcs[ibc, iface][6], data.elem[iel].bcs[ibc, iface][7]))

	if data.nbc < 2:
		outfile.write('  ***** NO THERMAL BOUNDARY CONDITIONS *****\n')
	outfile.write('    0 PRESOLVE/RESTART OPTIONS  *****\n')
	outfile.write('    7         INITIAL CONDITIONS *****\n')
	outfile.write(' C Default\n')
	outfile.write(' C Default\n')
	outfile.write(' C Default\n')
	outfile.write(' C Default\n')
	outfile.write(' C Default\n')
	outfile.write(' C Default\n')
	outfile.write(' C Default\n')
	outfile.write('  ***** DRIVE FORCE DATA ***** BODY FORCE, FLOW, Q\n')
	outfile.write('    4                 Lines of Drive force data follow\n')
	outfile.write(' C\n')
	outfile.write(' C\n')
	outfile.write(' C\n')
	outfile.write(' C\n')
	outfile.write(' ***** Variable Property Data ***** Overrrides Parameter data.\n')
	outfile.write('    1 Lines follow.\n')
	outfile.write('    0 PACKETS OF DATA FOLLOW\n')
	outfile.write(' ***** HISTORY AND INTEGRAL DATA *****\n')
	outfile.write('    0   POINTS.  Hcode, I,J,H,IEL\n')
	outfile.write(' ***** OUTPUT FIELD SPECIFICATION *****\n')
	outfile.write('    6 SPECIFICATIONS FOLLOW\n')
	outfile.write('    F      COORDINATES\n')
	outfile.write('    T      VELOCITY\n')
	outfile.write('    T      PRESSURE\n')
	outfile.write('    F      TEMPERATURE\n')
	outfile.write('    F      TEMPERATURE GRADIENT\n')
	outfile.write('    0      PASSIVE SCALARS\n')
	outfile.write(' ***** OBJECT SPECIFICATION *****\n')
	outfile.write('        0 Surface Objects\n')
	outfile.write('        0 Volume  Objects\n')
	outfile.write('        0 Edge    Objects\n')
	outfile.write('        0 Point   Objects\n')
	#
	# close file
	outfile.close()
	#
	# output
	return 0


def readre2(fname):
	"""A function for reading .re2 files for nek5000

	Parameters
	----------
	fname : str
		file name
	"""
	#
	try:
		infile = open(fname, 'rb')
	except IOError as e:
		logger.critical('I/O error ({0}): {1}'.format(e.errno, e.strerror))
		return -1
	# the header for re2 files is 80 ASCII bytes, something like
	# #v002    18669  2    18669 this is the hdr                                      %
	header = infile.read(80).split()
	nel = int(header[1])
	ndim = int(header[2])
	# always double precision
	wdsz = 8
	realtype = 'd'

	# detect endianness
	etagb = infile.read(4)
	etagL = struct.unpack('<f', etagb)[0]; etagL = int(etagL*1e5)/1e5
	etagB = struct.unpack('>f', etagb)[0]; etagB = int(etagB*1e5)/1e5
	if (etagL == 6.54321):
		logger.debug('Reading little-endian file\n')
		emode = '<'
		endian = 'little'
	elif (etagB == 6.54321):
		logger.debug('Reading big-endian file\n')
		emode = '>'
		endian = 'big'
	else:
		logger.error('Could not interpret endianness')
		return -3
	#
	# there are no GLL points here, only quad/hex vertices
	lr1 = [2, 2, ndim-1]
	npel = 2**ndim
	# the file only contains geometry
	var = [ndim, 0, 0, 0, 0]
	# allocate structure
	data = exdat.exadata(ndim, nel, lr1, var, 1)
	#
	# some metadata
	data.wdsz = wdsz
	data.endian = endian
	#
	# read the whole geometry into a buffer
	# for some reason each element is prefixed with 8 bytes of zeros, it's not clear to me why.
	# This is the reason for the +1 here, then the first number is ignored.
	buf = infile.read((ndim*npel+1)*wdsz*nel)
	elem_shape = [ndim, ndim-1, 2, 2]  # nvar, lz, ly, lx
	for (iel, el) in enumerate(data.elem):
		fi = np.frombuffer(buf, dtype=emode+realtype, count=ndim*npel+1, offset=(ndim*npel+1)*wdsz*iel)
		# the data is stored in the following order (2D|3D):
		# x1, x2, x4, x3; | x5, x6, x8, x7;
		# y1, y2, y4, y3; | y5, y6, y8, y7;
		# ----------------
		# z1, z2, z4, z3; z5, z6, z8, z7;
		# where 1-8 is the ordering of the points in memory
		for idim in range(ndim):  # x, y, [z]
			for iz in range(ndim-1):  # this does only one iteration in 2D, and in 3D does one iteration for 1-4 and one for 5-8
				el.pos[idim, iz, 0, 0] = fi[npel*idim+4*iz+1]
				el.pos[idim, iz, 0, 1] = fi[npel*idim+4*iz+2]
				el.pos[idim, iz, 1, 1] = fi[npel*idim+4*iz+3]
				el.pos[idim, iz, 1, 0] = fi[npel*idim+4*iz+4]
	#
	# read curved sides
	# the number of curved sides is stored as a double,
	# then each curved side is 64 bytes:
	# iel iface p1 p2 p3 p4 p5 ctype
	# where p1-5 are f64 parameters and ctype is the type of curvature in ASCII
	ncparam = 8
	buf = infile.read(wdsz)
	ncurv = int(np.frombuffer(buf)[0])
	logger.debug('Found {} curved sides'.format(ncurv))
	data.ncurv = ncurv
	buf = infile.read(wdsz*(ncparam*ncurv))
	for icurv in range(ncurv):
		# interpret the data
		curv = np.frombuffer(buf, dtype=emode+realtype, count=ncparam, offset=icurv*ncparam*wdsz)
		iel = int(curv[0])-1
		iedge = int(curv[1])-1
		cparams = curv[2:7]
		# select only the first byte, because it turns out the later bytes may contain garbage.
		# typically, it's b'C\x00\x00\x00\x00\x00\x00\x00' or b'C\x00\x00\x00\x00\x00\xe0?'.
		# AFAIK the curvature types are always one character long anyway.
		ctype = curv[7].tobytes()[:1].decode('utf-8')
		# fill in the data structure
		data.elem[iel].curv[iedge, :] = cparams
		data.elem[iel].ccurv[iedge] = ctype
	#
	# read boundary conditions
	# there can be more than one field, and we won't know until we'vre reached the end
	nbcparam = 8
	buf = infile.read(wdsz)
	ifield = 0
	while buf != b'':
		# the data is initialized with one BC field, we might need to allocate another
		if ifield > 0:
			data.nbc = data.nbc+1
			for el in data.elem:
				empty_bcs = np.zeros(el.bcs[:1, :].shape, dtype=el.bcs.dtype)
				el.bcs = np.concatenate((el.bcs, empty_bcs))
		nbclines = int(np.frombuffer(buf)[0])
		logger.debug('Found {} external boundary conditions for field {}'.format(nbclines, ifield))
		buf = infile.read(wdsz*(nbcparam*nbclines))
		for ibc in range(nbclines):
			# interpret the data
			bc = np.frombuffer(buf, dtype=emode+realtype, count=nbcparam, offset=ibc*nbcparam*wdsz)
			iel = int(bc[0])-1
			iface = int(bc[1])-1
			bcparams = bc[2:7]
			bctype = bc[7].tobytes().decode('utf-8').rstrip()  # remove trailing spaces
			# fill in the data structure
			data.elem[iel].bcs[ifield, iface][0] = bctype
			data.elem[iel].bcs[ifield, iface][1] = iel+1
			data.elem[iel].bcs[ifield, iface][2] = iface+1
			for ipar in range(5):
				data.elem[iel].bcs[ifield, iface][3+ipar] = bcparams[ipar]
		ifield = ifield + 1
		# try reading the number of conditions in the next field
		buf = infile.read(wdsz)
	infile.close()
	return data


def writere2(fname, data):
	"""A function for writing binary .re2 files for nek5000

	Parameters
	----------
	fname : str
		file name
	data : exadata
		data organised as in exadata.py
	"""
	#
	#---------------------------------------------------------------------------
	# CHECK INPUT DATA
	#---------------------------------------------------------------------------
	#
	# We could extract the corners, but for now just return an error if lr1 is too large
	if data.lr1 != [2, 2, data.ndim-1]:
		logger.critical('wrong element dimensions for re2 file! {} != {}'.format(data.lr1, [2, 2, data.ndim-1]))
		return -2
	#
	if data.var[0] != data.ndim:
		logger.critical('wrong number of geometric variables for re2 file! expected {}, found {}'.format(data.ndim, data.var[0]))
		return -3
	#
	# Open file
	try:
		outfile = open(fname, 'wb')
	except IOError as e:
		logger.critical('I/O error ({0}): {1}'.format(e.errno, e.strerror))
		return -1
	#
	#---------------------------------------------------------------------------
	# WRITE HEADER
	#---------------------------------------------------------------------------
	#
	# always double precision
	wdsz = 8
	realtype = 'd'
	nel = data.nel
	ndim = data.ndim
	header = '#v002{:9d}{:3d}{:9d} this is the hdr'.format(nel, ndim, nel)
	header = header.ljust(80)
	outfile.write(header.encode('utf-8'))
	#
	# decide endianness
	if data.endian in ('big', 'little'):
		byteswap = data.endian != sys.byteorder
		logger.debug(f'Writing {data.endian}-endian file')
	else:
		byteswap = False
		logger.warning(
			f'Unrecognized endianness {data.endian}, '
			f'writing native {sys.byteorder}-endian file'
		)
	#
	def correct_endianness(a):
		''' Return the array with the requested endianness'''
		if byteswap:
			return a.byteswap()
		else:
			return a
	#
	# write tag (to specify endianness)
	endianbytes = np.array([6.54321], dtype=np.float32)
	correct_endianness(endianbytes).tofile(outfile)
	#
	#---------------------------------------------------------------------------
	# WRITE DATA
	#---------------------------------------------------------------------------
	#
	# compute total number of points per element
	npel = 2**ndim
	#
	def write_data_to_file(a):
		'''Write the geometry of an element to the output file in double precision'''
		correct_endianness(a).tofile(outfile)
	#
	# write geometry (adding eight bytes of zeros before each element)
	xyz = np.zeros((npel*ndim+1,))  # array storing reordered geometry data (with a zero in the first position)
	for el in data.elem:
		# the data is stored in the following order (2D|3D):
		# x1, x2, x4, x3; | x5, x6, x8, x7;
		# y1, y2, y4, y3; | y5, y6, y8, y7;
		# ----------------
		# z1, z2, z4, z3; z5, z6, z8, z7;
		# where 1-8 is the ordering of the points in memory
		for idim in range(ndim):  # x, y, [z]
			for iz in range(ndim-1):  # this does only one iteration in 2D, and in 3D does one iteration for 1-4 and one for 5-8
				xyz[npel*idim+4*iz+1] = el.pos[idim, iz, 0, 0]
				xyz[npel*idim+4*iz+2] = el.pos[idim, iz, 0, 1]
				xyz[npel*idim+4*iz+3] = el.pos[idim, iz, 1, 1]
				xyz[npel*idim+4*iz+4] = el.pos[idim, iz, 1, 0]
		write_data_to_file(xyz)
	#
	# write curve sides data
	# locate curved edges
	curved_edges = []
	for (iel, el) in enumerate(data.elem):
		for iedge in range(12):
			if el.ccurv[iedge] != '':
				curved_edges.append((iel, iedge))
	# write number of curved edges
	ncurv = len(curved_edges)
	if ncurv != data.ncurv:
		logger.warning('wrong number of curved edges: expected {}, found {}'.format(data.ncurv, ncurv))
	ncurvf = np.array([ncurv], dtype=np.float64)
	write_data_to_file(ncurvf)
	# format curve data
	cdata = np.zeros((ncurv,), dtype='f8, f8, f8, f8, f8, f8, f8, S8')
	for (cdat, (iel, iedge)) in zip(cdata, curved_edges):
		el = data.elem[iel]
		cdat[0] = iel+1
		cdat[1] = iedge+1
		# curve parameters
		for j in range(5):
			cdat[2+j] = el.curv[iedge, j]
		# encode the string as a byte array padded with spaces
		cdat[7] = el.ccurv[iedge].encode('utf-8')
	# write to file
	write_data_to_file(cdata)
	#
	# write boundary conditions for each field
	for ifield in range(data.nbc):
		# locate faces with boundary conditions
		bc_faces = []
		for (iel, el) in enumerate(data.elem):
			for iface in range(2*ndim):
				bctype = el.bcs[ifield, iface][0]
				# internal boundary conditions are not written to .re2 files by reatore2
				# and are apparently ignored by Nek5000 even in .rea files
				if bctype != '' and bctype != 'E':
					bc_faces.append((iel, iface))
		nbcs = len(bc_faces)
		nbcsf = np.array([nbcs], dtype=np.float64)
		write_data_to_file(nbcsf)
		# initialize and format data
		bcdata = np.zeros((nbcs,), dtype='f8, f8, f8, f8, f8, f8, f8, S8')
		for (bc, (iel, iface)) in zip(bcdata, bc_faces):
			el = data.elem[iel]
			bc[0] = iel+1
			bc[1] = iface+1
			for j in range(5):
				bc[2+j] = el.bcs[ifield, iface][3+j]
			# encode the string as a byte array padded with spaces
			bc[7] = el.bcs[ifield, iface][0].encode('utf-8').ljust(8)
		# write to file
		write_data_to_file(bcdata)
	#
	# close file
	outfile.close()
	# rerurn
	return 0
